import setuptools

setuptools.setup(
    name='pydictable',
    version='0.9.5',
    author='Pramod Kumar',
    author_email='pramodkumar.damam73@gmail.com',
    description='Make your classes from/to dict',
    url='https://github.com/pskd73/pydictable.git',
    packages=['pydictable'],
    include_package_data=True,
    long_description='A tool to create data modals from dict and convert it to dict. '
                     'Check documentation here https://github.com/pskd73/pydictable',
    long_description_content_type='text/plain'
)
